------------------------------------------------------------------------------
--  Copyright (c) 2018 by Paul Scherrer Institute, Switzerland
--  All rights reserved.
--  Project:      evr320
--  Authors:      Jonas Purtschert
--  Description:  Filter a specific data field from data buffer stream of the decoder:
--  Modif:        Benoit Stef 
--                not based on counter anymore but map to LSB into a generic array
------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.psi_common_math_pkg.all;

entity evr320_data_filter is
  generic (NUM_BYTES : integer  := 8;
           SWAP      : boolean  := false  );  -- if true byte are swpped at output
  port (   -- User stream interface
          i_stream_clk        : in    std_logic;                          -- user clock
          i_stream_data       : in    std_logic_vector(7 downto 0);  
          i_stream_addr       : in    std_logic_vector(10 downto 0);  
          i_stream_valid      : in    std_logic;
          i_address           : in    std_logic_vector(11 downto 0);
          -- filter output:
          o_data              : out   std_logic_vector(NUM_BYTES*8-1 downto 0) := (others=>'0');
          o_valid             : out   std_logic := '0'
        );
end entity;

architecture behavioral of evr320_data_filter is
  -- array to store data value prior to map to output  
  type byte_array_t is array (NUM_BYTES-1 downto 0) of std_logic_vector(i_stream_data'range);
  signal table_s    : byte_array_t;
  --pipe help 
  signal addr_dff_s : std_logic_vector(10 downto 0)         := (others=>'0');
  signal data_dff_s : std_logic_vector(i_stream_data'range) := (others=>'0');
  signal filt_dff_s : std_logic_vector(i_address'range);
  --helper
  constant low_bd_c : integer := log2ceil(NUM_BYTES);   --compute LSB for address decoding
  signal ena_s      : std_logic_vector(1 downto 0);     --clock enable vector
  
begin
  
  p : process(i_stream_clk)
  begin
    if (rising_edge(i_stream_clk)) then
      
       --*** 1st pipe stage ***
       ena_s(0) <= i_stream_valid;
       filt_dff_s <= i_address;
       if (i_stream_valid = '1') then
         addr_dff_s <= i_stream_addr;
         data_dff_s <= i_stream_data;
       end if;
      
      --*** filling the array ***
      if ena_s(0) = '1' then   
        if addr_dff_s(10 downto low_bd_c) = filt_dff_s(10 downto low_bd_c) then
            --*** spatial loop *** 
            ena_s(1) <= '1';
            for i in 0 to NUM_BYTES-1 loop
              if addr_dff_s(low_bd_c-1 downto 0) = to_uslv(i,low_bd_c) then
                table_s(i) <= data_dff_s ;
              end if;           
            end loop;
        end if;
      else 
        ena_s(1) <= '0';
      end if;
       
      --*** set the output & valid accordingly*** 
      if ena_s(1) = '1' then
        if (from_uslv(addr_dff_s) = (from_uslv(filt_dff_s) + NUM_BYTES -1))  then
          o_valid <= '1';
          --*** spatial loop map output ***
          for i in 0 to NUM_BYTES-1 loop
            if SWAP then
              o_data(NUM_BYTES+i*NUM_BYTES-1 downto 0+i*NUM_BYTES) <= table_s(NUM_BYTES-1-i);
            else
              o_data(NUM_BYTES+i*NUM_BYTES-1 downto 0+i*NUM_BYTES) <= table_s(i);
            end if;
          end loop;
        else
           o_valid <= '0';      
        end if;
      end if;
      
     end if;
  end process;
  
end architecture;
